# Copyright 2016 Evgeny Golyshev. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility code intended to simplify initializing the event loop. """

import asyncio

from tornado.platform.asyncio import AsyncIOMainLoop


class Singleton(type):
    """Metaclass which helps to create singletons. """
    instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls.instances:
            singleton = super(Singleton, cls)
            cls.instances[cls] = singleton.__call__(*args, **kwargs)

        return cls.instances[cls]


class IOLoop(metaclass=Singleton):  # pylint: disable=too-few-public-methods
    """A singleton intended to create the event loop based on the AsyncIO Event
    Loop.
    """

    def __init__(self):
        self._io_loop = None
        AsyncIOMainLoop().install()

    def start(self, app, port):
        """Starts the event loop. """

        app.listen(port)
        self._io_loop = asyncio.get_event_loop()
        self._io_loop.run_forever()



# Copyright 2016 Evgeny Golyshev. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module contains the implementation of the entity called request. When
an RPC client connects to an RPC server and invokes one of the available remote
procedures, the request is passed to the remote procedure. The request is
necessary to allow the remote procedure to "find" its client when returning the
result.
"""

from tornado.escape import json_encode


class Ret(Exception):
    """Exception raised when a remote procedure returns a value via
    `Request.ret`.
    """

    def __init__(self, value=None):
        super(Ret, self).__init__()
        self.value = value


class Request:
    """Base class for requests.

    The instance of the class is passed as the first argument to a remote
    procedure before invoking it.
    """

    def __init__(self, marker, callback):
        self._callback = callback
        self._marker = marker

    #
    # Internal methods
    #
    def _get_error_response(self, message):
        response = {
            'error': message,
            'marker': self._marker,
        }
        return json_encode(response)

    def _get_successful_response(self, result, eod=True):
        response = {
            'eod': 1 if eod else 0,
            'marker': self._marker,
            'result': result,
        }
        return json_encode(response)

    #
    # User visible methods
    #
    def ret(self, value):
        """Causes a remote procedure to exit and return the specified value to
        the RPC client. The return statement can be used instead.
        """
        self._callback(self._get_successful_response(value))
        raise Ret()

    def ret_and_continue(self, value):
        """Causes a remote procedure to return the specified value to the RPC
        client. Unlike ret, the method doesn't cause the procedure to exit.
        """
        self._callback(self._get_successful_response(value, False))

    def ret_error(self, message):
        """Causes a remote procedure to exit and inform the the client that an
        error occurred.
        """
        self._callback(self._get_error_response(message))
        raise Ret()



# Copyright 2016 Evgeny Golyshev. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module provides an RPC framework the primary goal of which is to
simplify creating microservices using Tornado.
"""

import logging
from functools import wraps

import jwt
import jwt.exceptions
import redis
import tornado
from tornado import gen
from tornado.escape import json_decode
from tornado.ioloop import IOLoop
from tornado.options import define, options
from tornado.platform.asyncio import to_asyncio_future
from tornado.websocket import WebSocketHandler

from shirow.request import Ret, Request

MOCK_TOKEN = 'mock_token'
MOCK_USER_ID = -1
TOKEN_PATTERN = r'([_\-\w\.]+)'

define('allow_mock_token',
       help="allow using '{}' instead of real token (for testing "
            "purposes only)".format(MOCK_TOKEN), default=False, type=bool)
define('config_file',
       help='load parameters from the specified configuration '
            'file', default='shirow.conf')
define('port',
       help='listen on a specific port', default=8888)
define('token_algorithm',
       help='specify the algorithm used to sign the token', default='HS256')
define('token_key',
       help='encrypt the token using the specified secret key', default=None)
define('redis_host',
       help='specify Redis host', default='localhost')
define('redis_port',
       help='specify Redis port', default=6379)


def check_number_of_args(method, params):
    """Checks if the number of actual arguments passed to a remote procedure
    matches the number of formal parameters of the remote procedure (except
    self and request).
    """

    min_args, max_args = method.arguments_range
    if (max_args - 2) >= len(params) >= (min_args - 2):
        return True

    return False


def remote(func):
    """Decorator to mark some of the methods of RPC servers as remote. The
    decorated methods can be considered as a part of the public interface,
    since they are accessible from the client side.
    """

    @wraps(func)
    @gen.coroutine
    def wrapper(self, *args, **kwargs):
        args += tuple(kwargs.values())
        return func(self, *args)

    try:
        defaults_number = len(func.__defaults__)
    # If there are no default arguments func.__defaults__ is None.
    except TypeError:
        defaults_number = 0

    arguments_number = func.__code__.co_argcount
    wrapper.arguments_range = (
        arguments_number - defaults_number,  # min
        arguments_number  # max
    )
    wrapper.remote = True

    return wrapper


class UndefinedMethod(Exception):
    """Exception raised when attempting to get a method which either does not
    exist or is not public."""


class RPCServer(WebSocketHandler):  # pylint: disable=abstract-method
    """Base class for RPC servers. """

    def __init__(self, application, request, **kwargs):
        WebSocketHandler.__init__(self, application, request, **kwargs)

        self.io_loop = IOLoop.current()
        self.logger = logging.getLogger('tornado.application')
        self.redis_conn = None
        self.user_id = None

    #
    # Internal methods.
    #
    @gen.coroutine
    def _call_remote_procedure(self, request, method_name, arguments_list):
        try:
            method = self._get_method(method_name)
        except UndefinedMethod:
            request.ret_error('the {} function is undefined'.
                              format(method_name))

        if check_number_of_args(method, arguments_list):
            future = to_asyncio_future(method(request, *arguments_list))
        else:
            request.ret_error('number of arguments mismatch in the {} '
                              'function call'.format(method_name))

        try:
            result = yield from future
            if result is not None:  # if the return statement was used
                request.ret(result)
        except Ret:
            pass
        except Exception:  # pylint: disable=broad-except
            message = 'an error occurred while executing the function'
            self.logger.exception(message)
            request.ret_error(message)

    def _decode_token(self, encoded_token):
        decoded = True
        try:
            token = jwt.decode(encoded_token, options.token_key,
                               algorithms=[options.token_algorithm])
        except jwt.exceptions.DecodeError:
            token = {}
            decoded = False

        self.user_id = token.get('user_id', None)
        return decoded

    def _dismiss_request(self):
        self.logger.warning('Authentication request was dismissed')
        self.set_header('WWW-Authenticate', 'Token realm="shirow"')
        self.set_status(401)  # Unauthorized
        self.finish()

    def _fail_request(self, message):
        self.logger.error(message)
        self.set_status(500)  # Internal Server Error
        self.finish()

    def _get_token_key(self, user_id=None):
        if user_id is None:
            user_id = self.user_id

        return 'user:{}:token'.format(user_id)

    def _get_method(self, method_name):
        method = getattr(self, method_name, None)
        if method_name in dir(self) and hasattr(method, 'remote'):
            return method

        raise UndefinedMethod

    def _open_redis_connection(self):
        if not self.redis_conn:
            self.redis_conn = redis.StrictRedis(host=options.redis_host,
                                                port=options.redis_port, db=0)
        try:
            connected = bool(self.redis_conn.ping())
        except redis.exceptions.ConnectionError:
            connected = False

        return connected

    @tornado.web.asynchronous
    def get(self, *args, **kwargs):
        try:
            encoded_token = args[0]
        except IndexError:  # The request doesn't contain a token.
            self._dismiss_request()
            return

        if options.token_key is None:
            self._fail_request('A token key must be specified either in the '
                               'configuration file or on the command line')
            return

        if not self._open_redis_connection():
            self._fail_request('Shirow is not able to connect to Redis')
            return

        # Prepare mock token if needed
        if options.allow_mock_token and encoded_token == MOCK_TOKEN:
            payload = {'user_id': MOCK_USER_ID, 'ip': '127.0.0.1'}
            encoded_token = \
                jwt.encode(payload, options.token_key,
                           algorithm=options.token_algorithm).decode('utf8')

            key = self._get_token_key(MOCK_USER_ID)
            token_ttl = 15  # seconds
            self.redis_conn.setex(key, 60 * token_ttl, encoded_token)
            self.user_id = MOCK_USER_ID

        decoded_token = self._decode_token(encoded_token)
        if not decoded_token:
            self._fail_request('An error occurred while decoding the '
                               'following token: {}'.format(encoded_token))
            return

        key = self._get_token_key()
        if self.redis_conn.get(key) == encoded_token.encode('utf8'):
            # The WebSocket connection request must not contain any parameters.
            # The only parameter we needed has already been processed. Now we
            # have to get rid of it.
            args = ()
            WebSocketHandler.get(self, *args, **kwargs)
        else:
            self._dismiss_request()

    def create(self):
        """Invoked when a connection to the RPC server is established. """

    def destroy(self):
        """Invoked when a connection to the RPC server is terminated. """

    # Implementing the methods inherited from
    # tornado.websocket.WebSocketHandler

    def log_exception(self, typ, value, tb):
        """Logs uncaught exceptions. This overrides the method in
        WebSocketHandler not to log exceptions derived from Ret.
        """
        if not isinstance(value, Ret):
            WebSocketHandler.log_exception(self, typ, value, tb)

    def open(self, *args, **kwargs):
        self.create()

    def on_close(self):
        self.destroy()

    @gen.coroutine
    def on_message(self, message):
        parsed = json_decode(message)

        def callback(response):
            self.write_message(response)

        request = Request(parsed['marker'], callback)

        method_name = parsed['function_name']
        params = parsed['parameters_list']

        try:
            yield self._call_remote_procedure(request, method_name, params)
        except Ret:
            pass



# Copyright 2018 Evgeny Golyshev. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Miscellaneous utility functions. """

from subprocess import PIPE

from tornado import gen
from tornado.process import Subprocess


@gen.coroutine
def execute_async(command_line, env=None):
    """Executes the specified command line asynchronously. """

    process = Subprocess(command_line, env=env, stdout=PIPE, stderr=PIPE)
    ret = yield process.wait_for_exit(raise_error=False)
    out, err = process.stdout.read(), process.stderr.read()  # pylint: disable=no-member
    process.stdout.close()
    process.stderr.close()
    return ret, out, err



import re
from setuptools import setup

try:
    import pypandoc
    with open('README.md', 'r') as f:
        long_description = re.sub('</?p[^>]*>', '', f.read())
        long_description = re.sub('<img[^>]*>', '', long_description)
        long_description = pypandoc.convert_text(long_description, 'rst',
                                                 format='md')
except ImportError:
    long_description = ('Shirow is an implementation of a distinctive concept '
                        'of a remote procedure call.')


setup(name='shirow',
      version='0.4',
      description='Shirow package',
      long_description=long_description,
      url='https://github.com/tolstoyevsky/shirow',
      author='CusDeb Team',
      maintainer='Evgeny Golyshev',
      maintainer_email='Evgeny Golyshev <eugulixes@gmail.com>',
      license='http://www.apache.org/licenses/LICENSE-2.0',
      packages=['shirow'],
      install_requires=[
          'tornado==4.5.3',
          'redis',
          'pyjwt'
      ])
